#!/usr/bin/perl -w
#
#  Project Manager, Take 9
#

my $libdir;
BEGIN {
    sub expandPath {
        my $x = shift;
        my $y = shift;
        if (-l "$x/$y") {
            my $l = readlink("$x/$y");
            if ($l =~ m|^(/.*)/|) {
                $1;
            } elsif ($l =~ m|(.*)/|) {
                "$x/$1"
            } else {
                $x;
            }
        } else {
            $x;
        }
    }

    if (exists $ENV{P9LIB}) {
        $libdir = $ENV{P9LIB}
    } else {
        my @paths;
        if ($0 =~ m|(.*)/(.*)|) {
            # Invoked with path.
            @paths = (expandPath($1,$2));
        } else {
            # Invoked through PATH search. Must include "." in search list, as we
            # might have been invoked as 'perl XXX'.
            my @pp = (".", split /:/, $ENV{PATH});
            foreach (@pp) { push @paths, expandPath($_, $0) }
        }
        foreach (@paths) {
            if (-d "$_/p9lib") { $libdir = "$_/p9lib"; last }
            if (-d "$_/../p9lib") { $libdir = "$_/../p9lib"; last }
        }
        if (!defined($libdir)) { die "Unable to locate library path, set \$P9LIB to override" }
    }
    push @INC, $libdir;
}

use strict;
use LoadedFile;
use ParsedFile;
use Expander;
use Util;
use Generator;

# Command Execution
sub run_x_parse(@);
sub run_x_expand(@);

# Command Line Parsing
sub cmd_main(@);

# Helpers
sub list_project_tree ( $$$ );

# Main Routine
$libdir = util_simplify_path($libdir);
while (@ARGV && $ARGV[0] =~ /^-{0,2}(.*?)=(.*)/) {
    cmd_set($1, $2);
    shift @ARGV;
}
cmd_init();
diag_info("init: library directory is '$libdir'");
cmd_main(@ARGV);
exit 0;

############################ Command Execution ############################

# x-parse FILE....
#   Read files, and print parsed representation
sub run_x_parse(@) {
    my $file = shift;
    my $pfile = LoadedFile->load($file);
    my $presult = ParsedFile->parse($pfile);
    $presult = util_clone($presult);

    foreach (@_) {
        my $pnew = ParsedFile->parse(LoadedFile->load($_));
        $presult->merge($pnew);
    }

    $presult->dump("");
}

# x-expand ....
#   Test variable expansion
sub run_x_expand(@) {
    if (!@_) {
        print <<EOF;
$0 x-expand file... [+NS...] [make=X] [exp=X] [var=val...] [file...] [\$(ref)]

file       definition file to load, e.g. P/Settings
+NS        add namespace for lookup, e.g. "+_foo.c"
make=0|1   disable/enable expansion of Make variables ("a=b")
exp=0|1    disable/enable expansion of expandable variables ("a:=b")
var=val    define local variables
\$(ref)     text to expand

Note that this command requires global variables (e.g. "dir=" or "verbose=")
to be set before the command verb, everything else will become local variables.
EOF
        return;
    }

    my $file = shift;
    my $presult = ParsedFile->parse(LoadedFile->load($file));

    my $e = Expander->new($presult);

    foreach (@_) {
        if (/^\+(.*)/) {
            diag_info "x-expand: add namespace $1";
            $e->addNamespace($1);
        } elsif (/^make=(.*)/) {
            diag_info "x-expand: set expand make $1";
            $e->setExpandMake($1);
        } elsif (/^exp=(.*)/) {
            diag_info "x-expand: set expand expandable $1";
            $e->setExpandExpandable($1);
        } elsif (/^(.*?)=(.*)/) {
            diag_info "x-expand: add local variable $1 = $2";
            $e->addLocalVariable($1, $2);
        } elsif (!/\$/) {
            diag_info "x-expand: load $_";
            $presult->merge(ParsedFile->parse(LoadedFile->load($_)));
        } else {
            print "$_:\n";
            print "  refs -> ", $e->expandRefs($_), "\n";
            print "  vals -> ", $e->expandVars($_), "\n";
        }
    }
}

# list OBJECT....
#    List objects of project. Possible OBJECTs include:
#    - files [all files mentioned in $(FILES_$(TARGETS))]
#    - projects [all projects mentioned in $(TARGETS)]
#    - dump [parsed representation of project]
#    - suffixes [which suffixes can be built from which input]
#    - makefiles [all makefiles mentioned in $(GENERATE)]
sub run_list(@) {
    my $p = proj_load();
    my $headings = cmd_bool('headings', @_ > 1);

my $list_help = "$0 list [headings=yes|no] ITEM...

  help            Help message
  files           File names
  projects        Project names
  dump            Dump full project contents
  makefiles       Names of generated files
  tree            Projects and their dependencies indented as tree
  project-matrix  Checkbox matrix of files/projects
  project-files   List of projects and their files\n";
    if (@_ == 0) { print $list_help; }

    foreach (@_) {
        if ($_ =~ /^-?-?help$/) {
            print "# help:\n" if $headings;
            print $list_help;
        } elsif ($_ eq 'files') {
            print "# files:\n" if $headings;
            list_files_to($p, \*STDOUT);
        } elsif ($_ eq 'projects') {
            print "# projects:\n" if $headings;
            foreach ($p->getVarListValue('TARGETS')) {
                print "$_\n";
            }
        } elsif ($_ eq 'dump') {
            print "# dump:\n" if $headings;
            $p->dump("");
        } elsif ($_ eq 'makefiles') {
            print "# makefiles:\n" if $headings;
            foreach ($p->getVarListValue('GENERATE')) {
                print "$_\n";
            }
        } elsif ($_ eq 'tree') {
            print "# tree:\n" if $headings;
            # Build index
            my %parent;
            foreach my $par ($p->getVarListValue('TARGETS')) {
                foreach my $sub ($p->getVarListValue("DEPEND_$par")) {
                    $parent{$sub} = $par;
                }
            }
            # Display it
            foreach my $par ($p->getVarListValue('TARGETS')) {
                if (!exists $parent{$par}) {
                    list_project_tree($p, '', $par);
                }
            }
        } elsif ($_ eq 'project-matrix') {
            print "# project-matrix:\n" if $headings;
            list_project_matrix($p);
        } elsif ($_ eq 'project-files') {
            print "# project-files:\n" if $headings;
            foreach my $proj ($p->getVarListValue('TARGETS')) {
                print "$proj:\n";
                foreach ($p->getVarListValue("FILES_$proj")) {
                    print "  $_\n";
                }
            }
        } else {
            diag_error("Unknown list object '$_'");
        }
    }
}

# update [FILE...]
#    Regenerate specified makefiles; all files if non specified
sub run_update {
    # Load project
    my $p = proj_load();

    # Find file list
    my $x = Expander->new($p);
    $x->setExpandExpandable(1);
    $x->setExpandMake(1);
    my @files = $p->getVarListValue('GENERATE');
    if (!@files) {
        diag_error("Project does not define files to generate (using the 'GENERATE' variable)");
    }

    # Generate
    if (@_) {
        foreach my $file (@_) {
            if (!grep {$_ eq $file} @files) {
                diag_error("Project does not define generated file '$file'");
            }
            proj_generate($p, $file);
        }
    } else {
        foreach my $file (@files) { proj_generate($p, $file); }
    }
}

# add-file FILE... [share=y|n] [proj=name] [force=y|n]
#    Add new file
sub run_add_file {
    my $p = proj_load();

    # Command-line and Project Settings
    my $optShare    = cmd_bool('share', $p->getVarStringValue('PROJ_SHARE'));
    my @optAuto     = $p->getVarListValue('PROJ_AUTO');
    my $optIgnoreRX = util_globs_to_regexp($p->getVarListValue('PROJ_IGNORE'),
                                           "*.o", "*.obj", "*~", "*.bak", "P/*", "CVS/*");
    my $optProject  = cmd_val('proj', '');

    # Build list of files
    my %fileProjects;
    foreach my $pr ($p->getVarListValue('TARGETS')) {
        foreach my $f ($p->getVarListValue("FILES_$pr")) {
            $fileProjects{$f} = $pr;
        }
    }

    my $didAuto = 0;
    foreach my $file (@_) {
        # Check for blacklist
        if (!cmd_bool('force')) {
            if ("/$file" =~ m|/$optIgnoreRX$|) {
                diag_warning("add-file: '$file' is on blacklist, ignoring it");
                next;
            }
        }

        # Figure out where to add it
        my $theProject = $optProject;
        if ($theProject eq '') {
            # Nothing given on command line, so check Auto list
            foreach (@optAuto) {
                if (/^(.+?):(.*)/) {
                    my $p = $1;
                    my $re = util_globs_to_regexp(split(/,/,$2));
                    if ($file =~ m|^$re$|) {
                        if ($theProject eq '') {
                            $theProject = $p
                        } elsif ($p ne $theProject) {
                            diag_warning("add-file: '$file' claimed by '$theProject' and '$p'");
                        }
                    }
                } else {
                    diag_warning("add-file: invalid entry '$_' in PROJ_AUTO")
                      unless $didAuto;
                }
            }
            $didAuto = 1;

            # Still nothing found?
            if ($theProject eq '') {
                $theProject = $p->getVarStringValue('PROJ_CURRENT');
            }
            if ($theProject eq '') {
                diag_warning("add-file: no project specified for '$file', ignoring it");
                next;
            }
        }

        # Does this project exist? We do NOT check whether it's a
        # member of TARGETS, we only need to know that it has a FILES
        # variable. This makes it easy to deal with temporarily
        # disabled projects.
        if (!defined($p->getVar("FILES_$theProject"))) {
            diag_warning("add-file: project '$theProject' specified for '$file' does not exist, ignoring file");
            next;
        }

        # Check for sharing
        if (!$optShare && exists $fileProjects{$file}) {
            my $warnProject = $fileProjects{$file};
            if (grep {$_ eq $file} $p->getVarListValue("FILES_$theProject")) {
                $warnProject = $theProject;
            }
            diag_warning("add-file: '$file' already in '$warnProject'");
            next;
        }

        # Although we already checked for dupes above, we still have
        # to check once more, since the project might not be in
        # TARGETS, and we don't want to have dupes in a project either
        # even if sharing is allowed
        if (grep {$_ eq $file} $p->getVarListValue("FILES_$theProject")) {
            diag_warning("add-file: '$file' already in '$theProject'");
            next;
        }

        # Finally, add the file.
        $p->addToVarUser("FILES_$theProject", $file, 1);
    }
    $p = proj_save($p);
}

# remove-file FILE... [proj=name]
#     Remove file from project
sub run_remove_file {
    my $p = proj_load();
    my $optProj = cmd_val('proj','');
    my @projects = $p->getVarListValue('TARGETS');

    foreach my $file (@_) {
        if ($optProj ne '') {
            if (!remove_file_from_project($p, $optProj, $file)) {
                diag_warning("remove-file: '$file' not in '$optProj'");
            }
        } else {
            my $did = 0;
            foreach (@projects) {
                if (remove_file_from_project($p, $_, $file)) { $did = 1 }
            }
            if (!$did) {
                diag_warning("remove-file: '$file' not in any project");
            }
        }
    }

    $p = proj_save($p);
}

# validate
#     Perform some checks
sub run_validate {
    if (@_) { diag_warning("validate: no parameters expected"); }

    my $p = proj_load();

    # Things proj6 checks:
    # - every TARGET must have a TYPE_xxx
    # - projects mentioned in DEPEND_xxx exist, projects do not
    #   depend on themselves (circularity not checked)
    # - all TYPE_xxx should be mentioned in TARGETS
    # - all FILES_xxx should have a TYPE_xxx
    # - OUTPUT (nay, GENERATE) should be set
    # - reserved variables (OBJECTS_xxx) should not be set
    # - PROJ_CURRENT, if set, must be valid
    # - files in P (P9) shouldn't be Perl scripts
}

# xargs/grep/egrep/fgrep/ls/wc [ARGS...]
#     Apply specified command to all files in project
#     FIXME: proj6 applies files in sorted order?
sub run_command {
    my $p = proj_load();
    # FIXME: proper quoting!
    open PIPE, "|-", "xargs " . join(' ',@_)
    # open PIPE, "|-", "xargs", @_
      or diag_perror("$_[0]");
    list_files_to($p, \*PIPE);
    close PIPE;
}

# init [KEYWORDS...]
#     Initialize project. If already initialized, and force=yes, update
#     templates.
sub run_init {
    my $dir = cmd_val('dir');
    if (-f "$dir/Settings") {
        # Allow?
        if (!cmd_bool('force')) {
            diag_error("This directory has already been initialized. Use 'force=yes' to force update");
        }
        # Okay
        my @todo;
        diag_info("init: scanning...");
        opendir DIR, $dir or diag_perror($dir);
        while (defined (my $d = readdir(DIR))) {
            my $pfile = "$dir/$d";
            my $tfile = "$libdir/rules/$d";
            # Skip over some names
            next if $d =~ /^\./;
            next if $d =~ /\.pr$/;
            next if $d =~ /^Settings/;
            next if $d =~ /\.tmpl$/;
            # Is it a file
            if (!-d $pfile && !-l $pfile && -f $tfile) {
                # Update it
                push @todo, $d;
            }
        }
        closedir DIR;
        # Now copy all files:
        init_copy_files("$libdir/rules", $dir, @todo);
    } else {
        # Scan for files to copy and to instantiate
        my $srcdir = "$libdir/rules";
        my @toCopy;
        my @toInstantiate;
        my %instVars = ( GENERATE => "" );
        diag_info("init: scanning...");
        opendir DIR, $srcdir or diag_perror($srcdir);
        while (defined (my $d = readdir(DIR))) {
            my $srcfile = "$srcdir/$d";
            # Skip over some names
            next if $d =~ /^\./;
            next if $d =~ /\.pr$/;
            next if $d =~ /[\#~]$/;
            next if -d $srcfile;

            if ($d =~ /\.tmpl$/) {
                push @toInstantiate, $d;
            } else {
                my $pFile = ParsedFile->parse(LoadedFile->load("$srcfile", 1));
                if (@_) {
                    # User gave some keywords. Check whether we have those.
                    my $haveKeyword = 0;
                    foreach my $keyword ($pFile->getVarListValue('INIT_KEYWORDS')) {
                        if (grep {$_ eq $keyword} @_) { $haveKeyword = 1; last }
                    }
                    if (!$haveKeyword) {
                        diag_info("init: skipping '$d'");
                        next;
                    }
                }
                foreach my $gen ($pFile->getVarListValue('INIT_DEFAULT_OUTPUT')) {
                    if (!exists $instVars{"INCLUDE_$gen"}) {
                        $instVars{"INCLUDE_$gen"} = $d;
                        if ($instVars{GENERATE} ne '') { $instVars{GENERATE} .= ' ' }
                        $instVars{GENERATE} .= $gen;
                        diag_info("init: will generate '$gen' from '$d'");
                    }
                }
                push @toCopy, $d;
            }
        }
        closedir DIR;

        # Initialize anew
        mkdir $dir, 0777;
        init_copy_files($srcdir, $dir, @toCopy);

        # Instantiate
        if (@toInstantiate && cmd_bool('verbose')) {
            foreach (sort keys %instVars) {
                diag_info("init:   $_ = $instVars{$_}");
            }
        }
        foreach (@toInstantiate) {
            my $instance = $_;
            $instance =~ s/\.tmpl$//;
            diag_info("init: instantiating '$_'");
            init_instantiate_file("$srcdir/$_", "$dir/$instance", \%instVars);
        }
    }
}

################################# Helpers #################################

sub remove_file_from_project {
    my $p = shift;
    my $projName = shift;
    my $fileName = shift;

    # Fetch variable
    my $fileVar = $p->getVar("FILES_$projName");
    if (!defined($fileVar)) { return 0 }

    # Check each part of the value
    my $did = 0;
    foreach my $ele (@{$fileVar->{value}}) {
        my @s = split /\s+/, $ele->getValue();
        if (grep{$_ eq $fileName} @s) {
            $ele->setValue(join(' ', grep {$_ ne $fileName} @s));
            $ele->setModified();
            $did = 1;
        }
    }
    return $did;
}

sub list_files_to {
    my $p = shift;
    my $fd = shift;

    my %files;
    foreach my $target ($p->getVarListValue('TARGETS')) {
        foreach my $file ($p->getVarListValue("FILES_$target")) {
            print $fd "$file\n"
              unless exists($files{$file});
            $files{$file} = 1;
        }
    }
}

sub list_project_tree ( $$$ ) {
    my ($p, $prefix, $project) = @_;

    my $type = $p->getVarStringValue("TYPE_$project") || '<error:no-type>';
    print "$prefix$project ($type)\n";
    foreach ($p->getVarListValue("DEPEND_$project")) {
        list_project_tree($p, $prefix."  ", $_);
    }
}

sub list_project_matrix {
    my $p = shift;

    # Compute headings
    my @headings = sort $p->getVarListValue('TARGETS');
    my $width = 3;
    foreach (@headings) { if (length($_) > $width) { $width = length($_) } }
    ++$width;

    # Now, generate a line for each file
    my %files;
    for (my $index = 0; $index < @headings; ++$index) {
        foreach my $thisFile ($p->getVarListValue("FILES_" . $headings[$index])) {
            $files{$thisFile}[$index] = 1;
        }
    }

    # Print headings
    foreach (@headings) {
        my $pfx = ' ' x int(($width - length($_)) / 2);
        my $h = $pfx . $_;
        printf "%-${width}s", $h;
    }
    print "\n";

    foreach (@headings) {
        print '-' x ($width-1), " ";
    }
    print "\n";

    # Print files
    my $pfx = ' ' x int(($width - 3) / 2);
    foreach my $f (sort keys %files) {
        for (my $i = 0; $i < @headings; ++$i) {
            if (exists $files{$f}[$i]) {
                printf "%-${width}s", "$pfx"."[x]";
            } else {
                printf "%-${width}s", "$pfx"."[ ]";
            }
        }
        print " $f\n";
    }
}

sub init_copy_files {
    my $indir = shift;
    my $outdir = shift;
    if (!@_) {
        diag_info("init: nothing to copy");
    } else {
        foreach my $file (@_) {
            diag_info("init: copying '$file'");
            open IN, "< $indir/$file" or diag_perror("$indir/$file");
            open OUT, "> $outdir/$file" or diag_perror("$outdir/$file");
            print OUT <IN>;
            close IN;
            close OUT;
        }
    }
}

sub init_instantiate_file {
    my $infile = shift;
    my $outfile = shift;
    my $defs = shift;
    open IN, "< $infile" or diag_perror($infile);
    open OUT, "> $outfile" or diag_perror($outfile);
    while (<IN>) {
        if (/^\#\$(.*)/) {
            my $rx = util_globs_to_regexp($1);
            foreach (sort grep /^$rx$/, keys %$defs) {
                print OUT util_format_variable($_, "=", $defs->{$_});
            }
        } else {
            print OUT $_;
        }
    }
}

########################### Command Line Parsing ##########################

# Main loop.
sub cmd_main(@) {
    if (@_ == 0) { diag_error ("need to specify a command. Try `$0 --help' for help") }

    my $command = shift @_;
    $command =~ s/^--?//;

    # Commands proj4 has:             status     does
    # - init                          OK         make blank P directory
    # - add-type NAME WILDCARD...     never      add file type (concept does not exist here)
    # - rem(ove)-type                 never      remove file type
    # - get-(file-)type               never      guess file type from name
    # - add-gen(erated) TPL FILE      ?          add generated file; would update GENERATE
    # - add-file FILE...              OK         add source file
    # - rem(ove)-file                 ?          remove source file
    # - add-(sub)proj(ect) NAME DEP   ?          add subproject
    # - rem(ove)-(sub)proj(ect)       ?          remove subproject
    # - set-project                   ?          set PROJ_CURRENT
    # - list                          OK         implemented, but different
    # - update                        OK         regenerate everything
    # - show                          diff       integrated in 'list'
    # - rename / mv                   ?          rename files, also change #include and rm/add in CVS
    # - subst / sed OLD NEW OLD NEW   ?          global regexp replace
    # - add-rule OUT IN COMMANDS      ?          add new Make rule
    # - rem(ove)-rule                 ?          remove Make rule
    # - read                          ?          read back 'show' output
    # - verify                        missing    validate project definition
    # - edit                          ?          start editor on 'show' output, then 'read' it
    # - (e)grep                       OK         invoke grep on all files
    # proj6 adds (but doesn't have all of the above):
    # - fgrep / wc / ls / xargs       OK         similar to egrep
    # proj9 adds:
    # - x-parse, x-expand             OK         debugging

    if ($command eq 'help') {
        print <<EOF;
$0 -- A Project Manager and Makefile Maker

$0 [param=value...] command [args...]

Parameters:
  root=DIR        project root [.]
  dir=DIR         project data directory [root/P]
  verbose=yes|no  some more logging [no]
  quiet=yes|no    extra quiet [no]

Commands:
  help            this help screen

  update [FILE]   regenerate makefiles
  init KEYWDS...  initialize for new project
  add-file FN...  add new files to project
  remove-file FN  remove files from project

  list OBJ...     display various information
  grep ARGS...    grep
  ls|wc ARGS...   apply ls/wc to all files in project
  xargs CMD...    apply command to all files

  x-parse FILE    test makefile parsing
  x-expand [...]  test expander

(c) 2007 Stefan Reuther <streu\@gmx.de>
EOF
        exit 0;
    } elsif ($command eq 'x-parse') {
        run_x_parse(cmd_handle_args(@_));
    } elsif ($command eq 'x-expand') {
        run_x_expand(@_);
    } elsif ($command eq 'init') {
      run_init(cmd_handle_args(@_));
    } elsif ($command eq 'list') {
        run_list(cmd_handle_args(@_));
    } elsif ($command eq 'update' || $command eq 'generate') {
        run_update(cmd_handle_args(@_));
    } elsif ($command eq 'add-file') {
        run_add_file(cmd_handle_args(@_));
    } elsif ($command eq 'remove-file') {
        run_remove_file(cmd_handle_args(@_));
    } elsif ($command eq 'validate') {
        run_validate(cmd_handle_args(@_));
    } elsif ($command =~ '[ef]?grep' || $command eq 'ls' || $command eq 'wc') {
        run_command($command, @_);
    } elsif ($command eq 'xargs') {
        run_command(@_);
    } else {
        diag_error ("unknown comnand `$command'. Try `$0 --help' for help");
    }
}

############################### Project I/O ###############################

sub proj_load() {
    my $dir = cmd_val('dir','.');
    my $pfile = LoadedFile->load($dir . '/Settings');
    my $ptree = ParsedFile->parse($pfile, 1);

    # Load .pr files
    my @subfiles = ();
    my @skipfiles = $ptree->getVarListValue('PROJ_IGNORE_PROJECTS');
    push @skipfiles, split /\s+/, cmd_val('ignore', '');
    my $skiprx = util_globs_to_regexp(@skipfiles);

    if (!opendir(DIR, $dir)) { diag_perror($dir) }
    while (defined (my $entry = readdir(DIR))) {
        if ($entry =~ /\.pr$/) {
            push @subfiles, $entry
              unless $entry =~ /^$skiprx$/;
        }
    }
    closedir DIR;

    foreach (sort @subfiles) {
        my $psub = LoadedFile->load("$dir/$_");
        $ptree->merge(ParsedFile->parse($psub, 1));
    }

    $ptree;
}

sub proj_save( $ ) {
    my $pTree = shift;
    my $reload = $pTree->isReloadRequired();
    my $pList = $pTree->performUpdates();

    if (!cmd_bool('dry', 'n')) {
        foreach my $file (@$pList) {
            diag_info("update: updating '" . $file->getName() . "'...");
            $file->write($file->getName()); # already includes root
        }
        if (@$pList) {
            if ($reload) {
                diag_info("update: reloading project...");
                $pTree = proj_load();
            }
            diag_info("update: regenerating makefiles because of changes...");
            proj_update_everything($pTree);
        }
    } else {
        foreach my $file (@$pList) {
            diag_info("update: would update '" . $file->getName() . "'");
        }
        if ($reload) {
            diag_info("update: would reload project before regenerating.");
        }
    }
    $pTree;
}

sub proj_update_everything ( $ ) {
    my $p = shift;

    # Find file list
    my @files = $p->getVarListValue('GENERATE');
    if (!@files) {
        diag_warning("Project does not define files to generate (using the 'GENERATE' variable)");
    }

    # Generate
    foreach my $file (@files) { proj_generate($p, $file); }
}

sub instantiateRecursively( $$$$ );

sub instantiateRecursively( $$$$ ) {
    my $p = shift;
    my $thisType = shift;
    my $pDidParents = shift;
    my $e = shift;

    my $thisProject = $e->expandVars('$(NAME)');

    # Pre-pass. Instantiate all regular assignments. They will be instantiated
    # again in post-pass. This way, parents may reference variables defined by
    # us. Do not instantiate append-assignments, as those would otherwise be
    # processed twice.
    $p->instantiateAssignments($thisType, $e);

    # Process our parents
    my @parents = $thisType->getVarListValue('PARENT');
    foreach (@parents) {
        my $thisParent = shift @parents;
        if (!exists $pDidParents->{$thisParent}) {
            $pDidParents->{$thisParent} = 1;
            if (!exists $p->{sub}{$thisParent}) {
                diag_error("Project '$thisProject' references unknown parent '$thisParent'");
            }
            my $thisP = $p->{sub}{$thisParent};
            instantiateRecursively($p, $thisP, $pDidParents, $e);
            $p->addToVar("PROJECTS_USE_$thisParent", 1, Value->new($thisProject));
        }
    }

    # Post-pass, i.e. instantiate ourselves
    $p->instantiateFile($thisType, $e);
}

sub proj_generate( $$ ) {
    my $p = shift;
    my $name = shift;

    diag_info("update: generating '$name'...");

    # We need our own private copy of the project tree as we're going
    # to manipulate it.
    $p = $p->clone();

    # Load all additional files
    my $expander = Expander->new($p);
    $expander->addNamespace("_$name");
    $expander->setExpandExpandable(1);
    $expander->setExpandMake(1);
    my $files = $expander->expandVars('$(INCLUDE)');
    # FIXME: if no INCLUDE specified, this yields "$(INCLUDE)" and a bogus error
    foreach (split /\s+/, $files) {
        my $pfile = LoadedFile->load(cmd_val('dir', '.') . '/' . $_);
        my $ptree = ParsedFile->parse($pfile, 1);
        $p->merge($ptree);
    }

    # If there is a matching template, include that as well
    if (exists $p->{sub}{$name}) {
        $p->instantiateFile($p->{sub}{$name}, $expander);
    }

    # Build list of all files
    # FIXME: above, we use Expander. We probably should use it
    # consistently, to have namespace semantics everywhere, but
    # getList is just simpler now, and that's what is used
    # in 'list files' etc.
    my %allInputFiles;
    foreach ($p->getVarListValue('TARGETS')) {
        foreach ($p->getVarListValue("FILES_$_")) {
            $allInputFiles{$_} = 1;
        }
    }

    # Instantiate files
    my $pProductions = $p->processFiles([sort keys %allInputFiles]);

    # Instantiate targets
    my @targets = $p->getVarListValue('TARGETS');
    my %doneTargets;
    while (@targets) {
        # TODO: precondition processing
        my $thisProject = shift @targets;
        my $thisTypeName = $p->getVarStringValue("TYPE_$thisProject");
        if (!exists $p->{sub}{$thisTypeName}) {
            diag_error("Project '$thisProject' has unknown type '$thisTypeName'");
        }
        my $thisType = $p->{sub}{$thisTypeName};
        my $e = Expander->new($p);
        $e->addLocalVariable('NAME', $thisProject);
        $e->addNamespace("_$thisProject");

        # FIXME: Right now, we instantiate everything into the main
        # tree directly. A possibly better way would be to instantiate
        # everything into a temporary tree, and finally instantiate
        # that. The current implementation accesses OBJECT_EXT in a
        # weird way.

        # We cannot populate OBJECTS_xx now, as we may not yet know
        # OBJECT_EXT here. Set it to a temporary value so that
        # $(OBJECTS) is correctly expanded during instantiation.
        $p->setVar("OBJECTS_$thisProject", 0, Value->new());
        $p->addToVar("PROJECTS_$thisTypeName", 1, Value->new($thisProject));
        $p->addToVar("PROJECTS_USE_$thisTypeName", 1, Value->new($thisProject));
        $p->addToVar("PROJECTS_ALL", 1, Value->new($thisProject));

        my $files0 = join('|', $p->getVarListValue("FILES_$thisProject"));

        # Instantiate all parents (original version). This instantiates
        # parents in breath-first order.
        #my @parents = $thisType->getVarListValue('PARENT');
        #my %didparents;
        #$didparents{$thisTypeName} = 1;
        #while (@parents) {
        #    my $thisParent = shift @parents;
        #    if (!exists $didparents{$thisParent}) {
        #        if (!exists $p->{sub}{$thisParent}) {
        #            diag_error("Project '$thisProject' references unknown parent '$thisParent'");
        #        }
        #        my $thisP = $p->{sub}{$thisParent};
        #        push @parents, $thisP->getVarListValue('PARENT');
        #        $p->addToVar("PROJECTS_USE_$thisParent", 1, Value->new($thisProject));
        #        $p->instantiateFile($thisP, $e);
        #    }
        #}
        #$p->instantiateFile($thisType, $e);

        # This instantiates recursively in depth-first order, and
        # allows for a "pre-pass".
        my %didparents;
        $didparents{$thisTypeName} = 1;
        instantiateRecursively($p, $thisType, \%didparents, $e);

        my @exts = $p->getVarListValue('OBJECT_EXT');
        my @files = $p->getVarListValue("FILES_$thisProject");
        my $files1 = join('|', @files);

        # If the project definition defined or changed the FILES variable, we
        # must instantiate the new files, and figure out new productions, to
        # be able to populate OBJECTS_xxx.
        if ($files0 ne $files1) {
            diag_info("Updating production list because FILES_$thisProject changed...");
            my %newInputFiles;
            foreach (@files) { $newInputFiles{$_} = 1 unless exists $allInputFiles{$_}; }
            my $pNewProductions = $p->processFiles([sort keys %newInputFiles]);
            foreach (keys %$pNewProductions) { $pProductions->{$_} = $pNewProductions->{$_} }
        }

        my $extRX = "\\.(" . join('|', @exts) . ")\$";
        my @out;
        foreach (@files) {
            foreach (keys %{$pProductions->{$_}}) {
                push @out, $_
                  if /$extRX/;
            }
        }

        $p->setVar("OBJECTS_$thisProject", 0, Value->new(join(' ', @out)));
    }

    # Build an 'all' rule
    my $allRule = $p->addRule('all');
    if (!@{$allRule->getDepend()}) {
        $allRule->addDepend('$(ALL_TARGETS)');
        my $phRule = $p->addRule('.PHONY');
        $phRule->addDepend('all');
    }

    # Build a '.SUFFIXES' rule
    $p->addRule('.SUFFIXES')->addDepend(map{".$_"} $p->getUsedSuffixes());

    # Build rules for subprojects
    foreach ($p->getVarListValue('TARGETS')) {
        if (!defined($p->getRule($_))) {
            # HACK: there might be a rule that depends on $(OUTPUT_xxx) which
            # might be the same as xxx. Check that first. A better interpretation
            # would be to expand all variables first.
            my $e = Expander->new($p);
            $e->setExpandMake(1); $e->setExpandExpandable(1);
            my $expanded = $e->expandVars("\$(OUTPUT_$_)");
            if ($expanded ne $_ && $expanded ne '') {
                $p->addRule($_)->addDepend("\$(OUTPUT_$_)");
                $p->addRule('.PHONY')->addDepend($_);
            }
        }
    }

    # Rename variables with strange names
    if (util_bool($p->getVarStringValue('FMT_SAFE_NAMES'))) {
        $p->safeRenameVariables();
    }

    # Now generate the Makefile:
    my $mf = Generator->new($p);

    # Syntax adaption
    if ((my $v = $p->getVarStringValue('FMT_INCLUDE')) ne '') {
        $mf->setInclude($v);
    }

    # Instantiate variables required by user:
    my @vars = $p->getVarListValue('MAIN_VARS');
    foreach (@vars) {
        $mf->outputVariable($_);
    }
    if (@vars) {
        $mf->addSpaceAfterVars();
    }

    # Build list of rules, and sort them. This will also instantiate
    # variables as needed.
    my @rules;
    push @rules, grep /^\./, sort keys %{$p->{rule}};
    push @rules, $p->getVarListValue('MAIN_TARGETS');
    push @rules, 'all';
    push @rules, sort keys %{$p->{rule}};
    $mf->outputRules(@rules);
    $mf->outputPatternRules();

    diag_info("update: writing '$name'...");
    $mf->outputToFile(cmd_val('root', '.') . '/' . $name);
}
